@@ -1,6 +1,6 @@
 #!/usr/bin/env python3
 
-# Copyright (C) 2017-2022 The btclib developers
+# Copyright (C) The btclib developers
 #
 # This file is part of btclib. It is subject to the license terms in the
 # LICENSE file found in the top-level directory of this distribution.
@@ -10,37 +10,193 @@
 
 """Elliptic Curve Digital Signature Algorithm (ECDSA).
 
-   Implementation according to SEC 1 v.2:
+Implementation according to SEC 1 v.2:
 
-   http://www.secg.org/sec1-v2.pdf
+http://www.secg.org/sec1-v2.pdf
 
-   specialized with bitcoin canonical 'lower-s' form
-   to avoid accepting malleable signatures.
+specialized with bitcoin canonical 'lower-s' form
+to avoid accepting malleable signatures.
 """
 
-import secrets
-from hashlib import sha256
-from typing import List, Optional, Tuple, Union
+from __future__ import annotations
 
-from btclib.alias import HashF, JacPoint, Octets, Point
-from btclib.ecc import libsecp256k1
-from btclib.ecc.curve import Curve, secp256k1
-from btclib.ecc.curve_group import _double_mult, _mult
-from btclib.ecc.der import Sig
-from btclib.ecc.number_theory import mod_inv
-from btclib.ecc.rfc6979 import _rfc6979_
+import contextlib
+import secrets
+from dataclasses import InitVar, dataclass
+from hashlib import sha256
+from io import BytesIO
+
+from btclib import var_bytes
+from btclib.alias import BinaryData, HashF, JacPoint, Octets, Point
+from btclib.ec import Curve, libsecp256k1, secp256k1
+from btclib.ec.curve_group import _double_mult, _mult
+from btclib.ecc.libsecp256k1 import ecdsa_sign_, ecdsa_verify_
+from btclib.ecc.rfc6979_nonce import _rfc6979_nonce_, challenge_
 from btclib.exceptions import BTClibRuntimeError, BTClibValueError
-from btclib.hashes import challenge_, reduce_to_hlen
+from btclib.hashes import reduce_to_hlen
+from btclib.number_theory import mod_inv
 from btclib.to_prv_key import PrvKey, int_from_prv_key
 from btclib.to_pub_key import Key, point_from_key, pub_keyinfo_from_key
-from btclib.utils import bytes_from_octets
+from btclib.utils import bytes_from_octets, bytesio_from_binarydata, hex_string
+
+_DER_SCALAR_MARKER = b"\x02"
+_DER_SIG_MARKER = b"\x30"
 
 
-def gen_keys(
-    prv_key: Optional[PrvKey] = None, ec: Curve = secp256k1
-) -> Tuple[int, Point]:
-    "Return a private/public (int, Point) key-pair."
+def _serialize_scalar(scalar: int) -> bytes:
+    # 'highest bit set' padding included here
+    scalar_size = scalar.bit_length() // 8 + 1
+    scalar_bytes = scalar.to_bytes(scalar_size, byteorder="big", signed=False)
+    return _DER_SCALAR_MARKER + var_bytes.serialize(scalar_bytes)
 
+
+def _deserialize_scalar(sig_data_stream: BytesIO) -> int:
+    marker = sig_data_stream.read(1)
+    if marker != _DER_SCALAR_MARKER:
+        err_msg = f"invalid value header: {marker.hex()}"
+        err_msg += f", instead of integer element {_DER_SCALAR_MARKER.hex()}"
+        raise BTClibValueError(err_msg)
+
+    r_bytes = var_bytes.parse(sig_data_stream, forbid_zero_size=True)
+    if r_bytes[0] == 0 and r_bytes[1] < 0x80:
+        raise BTClibValueError("invalid 'highest bit set' padding")
+    if r_bytes[0] >= 0x80:
+        raise BTClibValueError("invalid negative scalar")
+
+    return int.from_bytes(r_bytes, byteorder="big", signed=False)
+
+
+@dataclass(frozen=True)
+class Sig:
+    """ECDSA signature with strict ASN.1 DER serialization.
+
+    The original Bitcoin implementation used OpenSSL to verify
+    ECDSA signatures in ASN.1 DER representation.
+    However, OpenSSL does not do strict validation
+    (e.g. extra padding is ignored) and this changes the transaction
+    hash value, leading to transaction malleability.
+    This was fixed by BIP66, activated on block 363,724.
+
+    source:
+    https://github.com/bitcoin/bips/blob/master/bip-0066.mediawiki
+
+    BIP66 mandates a strict DER format:
+
+    Format:
+    [0x30] [data-size][0x02][r-size][r][0x02][s-size][s]
+
+    * 0x30: header byte to indicate compound structure
+    * data-size: 1-byte size descriptor of the following data
+    * 0x02: header byte indicating an integer
+    * r-size: 1-byte size descriptor of the r value that follows
+    * r: arbitrary-size big-endian r value.
+        It must use the shortest possible encoding for
+        a positive integers: no null bytes at the start,
+        except a single one when the next byte has its highest bit set
+        (to avoid being interpreted as a negative number)
+    * 0x02: header byte indicating an integer
+    * s-size: 1-byte size descriptor of the s value that follows
+    * s: arbitrary-size big-endian s value. Same rules as for r apply
+
+    There are 7 bytes of meta-data:
+
+    * compound header, compound size,
+    * value header, r-value size,
+    * value header, s-value size
+
+    The ECDSA signature (r, s) should be 64 bytes,
+    r and s being 32 bytes integers each;
+    however, integers in DER are signed,
+    so if the value being encoded is greater than 2^128,
+    a 33rd byte is added in front.
+    Bitcoin has a "low s" rule for the s value to be below ec.n,
+    but it is only a standardness rule miners are allowed to ignore.
+    Moreover, no such rule exists for r.
+    """
+
+    # 32 bytes scalar, 0 < r < ec.n (ec.n is the curve order)
+    r: int
+    # 32 bytes scalar, 0 < s < ec.n (ec.n is the curve order)
+    s: int
+    ec: Curve = secp256k1
+    check_validity: InitVar[bool] = True
+
+    def __post_init__(self, check_validity: bool) -> None:
+        if check_validity:
+            self.assert_valid()
+
+    def assert_valid(self) -> None:
+        # r is a scalar, fail if r is not in [1, n-1]
+        if not 0 < self.r < self.ec.n:
+            err_msg = "scalar r not in 1..n-1: "
+            err_msg += f"'{hex_string(self.r)}'" if self.r > 0xFFFFFFFF else f"{self.r}"
+            raise BTClibValueError(err_msg)
+
+        # ensure r is congruent to a valid x-coordinate
+        r = self.r
+        congruence_not_found = True
+        while congruence_not_found and r < self.ec.p:
+            try:
+                self.ec.y(r)
+                congruence_not_found = False
+            except BTClibValueError:
+                r += self.ec.n
+        if congruence_not_found:
+            err_msg = "r is not (congruent to) a valid x-coordinate: "
+            err_msg += f"'{hex_string(self.r)}'" if self.r > 0xFFFFFFFF else f"{self.r}"
+            raise BTClibValueError(err_msg)
+
+        # s is a scalar, fail if s is not in [1, n-1]
+        if not 0 < self.s < self.ec.n:
+            err_msg = "scalar s not in 1..n-1: "
+            err_msg += f"'{hex_string(self.s)}'" if self.s > 0xFFFFFFFF else f"{self.s}"
+            raise BTClibValueError(err_msg)
+
+    def serialize(self, check_validity: bool = True) -> bytes:
+        """Serialize an ECDSA signature to strict ASN.1 DER representation."""
+        if check_validity:
+            self.assert_valid()
+
+        out = _serialize_scalar(self.r)
+        out += _serialize_scalar(self.s)
+        return _DER_SIG_MARKER + var_bytes.serialize(out)
+
+    @classmethod
+    def parse(cls: type[Sig], data: BinaryData, check_validity: bool = True) -> Sig:
+        """Return a Sig by parsing binary data.
+
+        Deserialize a strict ASN.1 DER representation of an ECDSA
+        signature.
+        """
+        stream = bytesio_from_binarydata(data)
+        ec = secp256k1
+
+        # [0x30] [data-size][0x02][r-size][r][0x02][s-size][s]
+        marker = stream.read(1)
+        if marker != _DER_SIG_MARKER:
+            err_msg = f"invalid compound header: {marker.hex()}"
+            err_msg += f", instead of DER sequence tag {_DER_SIG_MARKER.hex()}"
+            raise BTClibValueError(err_msg)
+
+        # [data-size][0x02][r-size][r][0x02][s-size][s]
+        sig_data = var_bytes.parse(stream, forbid_zero_size=True)
+
+        # [0x02][r-size][r][0x02][s-size][s]
+        sig_data_substream = bytesio_from_binarydata(sig_data)
+        r = _deserialize_scalar(sig_data_substream)
+        s = _deserialize_scalar(sig_data_substream)
+
+        # to prevent malleability
+        # the sig_data_substream must have been consumed entirely
+        if sig_data_substream.read(1) != b"":
+            err_msg = "invalid DER sequence length"
+            raise BTClibValueError(err_msg)
+
+        return cls(r, s, ec, check_validity)
+
+
+def gen_keys(prv_key: PrvKey | None = None, ec: Curve = secp256k1) -> tuple[int, Point]:
+    """Return a private/public (int, Point) key-pair."""
     if prv_key is None:
         # q in the range [1, ec.n-1]
         q = 1 + secrets.randbelow(ec.n - 1)
@@ -56,9 +212,7 @@ def _sign_(c: int, q: int, nonce: int, lower_s: bool, ec: Curve) -> Sig:
     # Private function for testing purposes: it allows to explore all
     # possible value of the challenge c (for low-cardinality curves).
     # It assume that c is in [0, n-1], while q and nonce are in [1, n-1]
-
     # Steps numbering follows SEC 1 v.2 section 4.1.3
-
     KJ = _mult(nonce, ec.GJ, ec)  # 1
 
     # affine x_K-coordinate of K (field element)
@@ -84,17 +238,16 @@ def _sign_(c: int, q: int, nonce: int, lower_s: bool, ec: Curve) -> Sig:
 def sign_(
     msg_hash: Octets,
     prv_key: PrvKey,
-    nonce: Optional[PrvKey] = None,
+    nonce: PrvKey | None = None,
     lower_s: bool = True,
     ec: Curve = secp256k1,
     hf: HashF = sha256,
 ) -> Sig:
     """Sign a hf_len bytes message according to ECDSA signature algorithm.
 
-    If the deterministic nonce is not provided,
-    the RFC6979 specification is used.
+    If the deterministic nonce is not provided, the RFC6979
+    specification is used.
     """
-
     # the message msg_hash: a hf_len array
     hf_len = hf().digest_size
     msg_hash = bytes_from_octets(msg_hash, hf_len)
@@ -103,21 +256,21 @@ def sign_(
     # SEC 1 v.2 section 3.2.1
     q = int_from_prv_key(prv_key, ec)
 
+    if (
+        ec == secp256k1
+        and nonce is None  # FIXME secp256k1 manage nonce
+        and lower_s
+        and hf == sha256
+        and libsecp256k1.is_available()
+    ):
+        return Sig.parse(ecdsa_sign_(msg_hash, q))
+
     # the challenge
     c = challenge_(msg_hash, ec, hf)  # 4, 5
 
-    if (
-        ec == secp256k1
-        and nonce is None
-        and lower_s
-        and hf == sha256
-        and libsecp256k1.is_enabled()
-    ):
-        return Sig.parse(libsecp256k1.dsa.sign(msg_hash, q))
-
     # nonce: an integer in the range 1..n-1.
     if nonce is None:
-        nonce = _rfc6979_(c, q, ec, hf)  # 1
+        nonce = _rfc6979_nonce_(c, q, ec, hf)  # 1
     else:
         nonce = int_from_prv_key(nonce, ec)
 
@@ -128,7 +281,7 @@ def sign_(
 def sign(
     msg: Octets,
     prv_key: PrvKey,
-    nonce: Optional[PrvKey] = None,
+    nonce: PrvKey | None = None,
     lower_s: bool = True,
     ec: Curve = secp256k1,
     hf: HashF = sha256,
@@ -152,7 +305,6 @@ def sign(
 
     See https://tools.ietf.org/html/rfc6979#section-3.2
     """
-
     msg_hash = reduce_to_hlen(msg, hf)
     return sign_(msg_hash, prv_key, nonce, lower_s, ec, hf)
 
@@ -187,30 +339,27 @@ def _assert_as_valid_(
 def assert_as_valid_(
     msg_hash: Octets,
     key: Key,
-    sig: Union[Sig, Octets],
+    sig: Sig | Octets,
     lower_s: bool = True,
     hf: HashF = sha256,
 ) -> None:
     # Private function for test/dev purposes
     # It raises Errors, while verify should always return True or False
-
     if isinstance(sig, Sig):
         sig.assert_valid()
     else:
         sig = Sig.parse(sig)
 
-    c = challenge_(msg_hash, sig.ec, hf)  # 2, 3
-
-    Q = point_from_key(key, sig.ec)
-    QJ = Q[0], Q[1], 1
-
-    if libsecp256k1.is_enabled() and sig.ec == secp256k1 and lower_s and hf == sha256:
+    if sig.ec == secp256k1 and hf == sha256 and libsecp256k1.is_available():
         msg_hash_bytes = bytes_from_octets(msg_hash)
         pubkey_bytes = pub_keyinfo_from_key(key)[0]
-        if not libsecp256k1.dsa.verify(msg_hash_bytes, pubkey_bytes, sig.serialize()):
-            raise BTClibRuntimeError("signature verification failed")
+        if not ecdsa_verify_(msg_hash_bytes, pubkey_bytes, sig.serialize(), lower_s):
+            raise BTClibRuntimeError("libsecp256k1.ecdsa_verify_ failed")
         return
 
+    c = challenge_(msg_hash, sig.ec, hf)  # 2, 3
+    Q = point_from_key(key, sig.ec)
+    QJ = Q[0], Q[1], 1
     # second part delegated to helper function
     _assert_as_valid_(c, QJ, sig.r, sig.s, lower_s, sig.ec)
 
@@ -218,13 +367,12 @@ def assert_as_valid_(
 def assert_as_valid(
     msg: Octets,
     key: Key,
-    sig: Union[Sig, Octets],
+    sig: Sig | Octets,
     lower_s: bool = True,
     hf: HashF = sha256,
 ) -> None:
     # Private function for test/dev purposes
     # It raises Errors, while verify should always return True or False
-
     msg_hash = reduce_to_hlen(msg, hf)
     assert_as_valid_(msg_hash, key, sig, lower_s, hf)
 
@@ -232,12 +380,11 @@ def assert_as_valid(
 def verify_(
     msg_hash: Octets,
     key: Key,
-    sig: Union[Sig, Octets],
+    sig: Sig | Octets,
     lower_s: bool = True,
     hf: HashF = sha256,
 ) -> bool:
     """ECDSA signature verification (SEC 1 v.2 section 4.1.4)."""
-
     # all kind of Exceptions are catched because
     # verify must always return a bool
     try:
@@ -251,27 +398,26 @@ def verify_(
 def verify(
     msg: Octets,
     key: Key,
-    sig: Union[Sig, Octets],
+    sig: Sig | Octets,
     lower_s: bool = True,
     hf: HashF = sha256,
 ) -> bool:
     """ECDSA signature verification (SEC 1 v.2 section 4.1.4)."""
-
     msg_hash = reduce_to_hlen(msg, hf)
     return verify_(msg_hash, key, sig, lower_s, hf)
 
 
-# TODO: use _recover_pub_key_ to avoid code duplication
+# TODO use _recover_pub_key_ to avoid code duplication
 def _recover_pub_keys_(
     c: int, r: int, s: int, lower_s: bool, ec: Curve
-) -> List[JacPoint]:
+) -> list[JacPoint]:
     # Private function provided for testing purposes only.
 
     # precomputations
     r_1 = mod_inv(r, ec.n)
     r1s = r_1 * s % ec.n
     r1e = -r_1 * c % ec.n
-    keys: List[JacPoint] = []
+    keys: list[JacPoint] = []
     # r = K[0] % ec.n
     # if ec.n < K[0] < ec.p (likely when cofactor ec.cofactor > 1)
     # then both x_K=r and x_K=r+ec.n must be tested
@@ -279,17 +425,14 @@ def _recover_pub_keys_(
         # affine x_K-coordinate of K (field element)
         x_K = (r + j * ec.n) % ec.p  # 1.1
         # two possible y_K-coordinates, i.e. two possible keys for each cycle
-        try:
+        with contextlib.suppress(BTClibValueError, BTClibRuntimeError):
             # even root first for bitcoin message signing compatibility
             yodd = ec.y_even(x_K)
             KJ = x_K, yodd, 1  # 1.2, 1.3, and 1.4
             # 1.5 has been performed in the recover_pub_keys calling function
             QJ = _double_mult(r1s, KJ, r1e, ec.GJ, ec)  # 1.6.1
-            try:
+            with contextlib.suppress(BTClibValueError, BTClibRuntimeError):
                 _assert_as_valid_(c, QJ, r, s, lower_s, ec)  # 1.6.2
-            except (BTClibValueError, BTClibRuntimeError):
-                pass
-            else:
                 keys.append(QJ)  # 1.6.2
             KJ = x_K, ec.p - yodd, 1  # 1.6.3
             QJ = _double_mult(r1s, KJ, r1e, ec.GJ, ec)
@@ -299,20 +442,17 @@ def _recover_pub_keys_(
                 pass
             else:
                 keys.append(QJ)  # 1.6.2
-        except (BTClibValueError, BTClibRuntimeError):  # K is not a curve point
-            pass
     return keys
 
 
 def recover_pub_keys_(
-    msg_hash: Octets, sig: Union[Sig, Octets], lower_s: bool = True, hf: HashF = sha256
-) -> List[Point]:
+    msg_hash: Octets, sig: Sig | Octets, lower_s: bool = True, hf: HashF = sha256
+) -> list[Point]:
     """ECDSA public key recovery (SEC 1 v.2 section 4.1.6).
 
     See also:
-    https://crypto.stackexchange.com/questions/18105/how-does-recovering-the-public-key-from-an-ecdsa-signature-work/18106#18106
+    - https://crypto.stackexchange.com/questions/18105/how-does-recovering-the-public-key-from-an-ecdsa-signature-work/18106#18106
     """
-
     if isinstance(sig, Sig):
         sig.assert_valid()
     else:
@@ -329,14 +469,13 @@ def recover_pub_keys_(
 
 
 def recover_pub_keys(
-    msg: Octets, sig: Union[Sig, Octets], lower_s: bool = True, hf: HashF = sha256
-) -> List[Point]:
+    msg: Octets, sig: Sig | Octets, lower_s: bool = True, hf: HashF = sha256
+) -> list[Point]:
     """ECDSA public key recovery (SEC 1 v.2 section 4.1.6).
 
     See also:
-    https://crypto.stackexchange.com/questions/18105/how-does-recovering-the-public-key-from-an-ecdsa-signature-work/18106#18106
+    - https://crypto.stackexchange.com/questions/18105/how-does-recovering-the-public-key-from-an-ecdsa-signature-work/18106#18106
     """
-
     msg_hash = reduce_to_hlen(msg, hf)
     return recover_pub_keys_(msg_hash, sig, lower_s, hf)
 
@@ -370,16 +509,15 @@ def _recover_pub_key_(
 def recover_pub_key_(
     key_id: int,
     msg_hash: Octets,
-    sig: Union[Sig, Octets],
+    sig: Sig | Octets,
     lower_s: bool = True,
     hf: HashF = sha256,
 ) -> Point:
     """ECDSA public key recovery (SEC 1 v.2 section 4.1.6).
 
     See also:
-    https://crypto.stackexchange.com/questions/18105/how-does-recovering-the-public-key-from-an-ecdsa-signature-work/18106#18106
+    - https://crypto.stackexchange.com/questions/18105/how-does-recovering-the-public-key-from-an-ecdsa-signature-work/18106#18106
     """
-
     if isinstance(sig, Sig):
         sig.assert_valid()
     else:
@@ -398,27 +536,26 @@ def recover_pub_key_(
 def recover_pub_key(
     key_id: int,
     msg: Octets,
-    sig: Union[Sig, Octets],
+    sig: Sig | Octets,
     lower_s: bool = True,
     hf: HashF = sha256,
 ) -> Point:
     """ECDSA public key recovery (SEC 1 v.2 section 4.1.6).
 
     See also:
-    https://crypto.stackexchange.com/questions/18105/how-does-recovering-the-public-key-from-an-ecdsa-signature-work/18106#18106
+    - https://crypto.stackexchange.com/questions/18105/how-does-recovering-the-public-key-from-an-ecdsa-signature-work/18106#18106
     """
-
     msg_hash = reduce_to_hlen(msg, hf)
     return recover_pub_key_(key_id, msg_hash, sig, lower_s, hf)
 
 
 def crack_prv_key_(
     msg_hash1: Octets,
-    sig1: Union[Sig, Octets],
+    sig1: Sig | Octets,
     msg_hash2: Octets,
-    sig2: Union[Sig, Octets],
+    sig2: Sig | Octets,
     hf: HashF = sha256,
-) -> Tuple[int, int]:
+) -> tuple[int, int]:
 
     if isinstance(sig1, Sig):
         sig1.assert_valid()
@@ -448,11 +585,11 @@ def crack_prv_key_(
 
 def crack_prv_key(
     msg1: Octets,
-    sig1: Union[Sig, Octets],
+    sig1: Sig | Octets,
     msg2: Octets,
-    sig2: Union[Sig, Octets],
+    sig2: Sig | Octets,
     hf: HashF = sha256,
-) -> Tuple[int, int]:
+) -> tuple[int, int]:
 
     msg_hash1 = reduce_to_hlen(msg1, hf)
     msg_hash2 = reduce_to_hlen(msg2, hf)
