@@ -1,14 +1,24 @@
-# Configuration file for the Sphinx documentation builder.
+#!/usr/bin/env python3
+
+# Copyright (C) The btclib developers
 #
-# This file only contains a selection of the most common options. For a full
-# list see the documentation:
-# http://www.sphinx-doc.org/en/master/config
+# This file is part of btclib. It is subject to the license terms in the
+# LICENSE file found in the top-level directory of this distribution.
+#
+# No part of btclib including this file, may be copied, modified, propagated,
+# or distributed except according to the terms contained in the LICENSE file.
+
+"""Configuration file for the Sphinx documentation builder.
+
+This file only contains a selection of the most common options. For a
+full list see the documentation: www.sphinx-doc.org/en/master/config
+"""
 
 # -- Path setup --------------------------------------------------------------
+from __future__ import annotations
 
 import sys
 from os import path
-from typing import List
 
 # If extensions (or modules to document with autodoc) are in another directory,
 # add these directories to sys.path here. If the directory is relative to the
@@ -19,11 +29,11 @@ sys.path.insert(0, path.abspath("../../"))
 # -- Project information -----------------------------------------------------
 
 project = "btclib"
-copyright = "2017-2022 The btclib developers"
+project_copyright = "2017-2023 The btclib developers"
 author = "The btclib developers"
 
 # The full version, including alpha/beta/rc tags
-release = "2022.7.20"
+release = "2023.1.17"
 
 
 # -- General configuration ---------------------------------------------------
@@ -39,7 +49,7 @@ templates_path = ["_templates"]
 # List of patterns, relative to source directory, that match files and
 # directories to ignore when looking for source files.
 # This pattern also affects html_static_path and html_extra_path.
-exclude_patterns: List[str] = []
+exclude_patterns: list[str] = []
 
 
 # -- Options for HTML output -------------------------------------------------
@@ -52,6 +62,6 @@ html_theme = "alabaster"
 # Add any paths that contain custom static files (such as style sheets) here,
 # relative to this directory. They are copied after the builtin static files,
 # so a file named "default.css" will overwrite the builtin "default.css".
-html_static_path: List[str] = []
+html_static_path: list[str] = []
 
 master_doc = "index"
